from typing import Dict, Union

from fastapi import APIRouter
import pandas as pd
from pydantic import Field


from model_server.base.accessors import GenericImageDataAccessor, PatchStack
from model_server.base.pipelines.params import PipelineParams, PipelineRecord
from model_server.base.pipelines.segment import segment_pipeline
from model_server.base.pipelines.util import call_pipeline
from model_server.base.roiset import get_label_ids, RoiSet, RoiSetMetaParams, RoiSetExportParams
from model_server.base.session import session
from model_server.base.util import PipelineTrace

from model_server.base.models import Model, InstanceSegmentationModel, SemanticSegmentationModel

router = APIRouter(
    prefix='/pipelines',
    tags=['pipelines'],
)

class RoiSetObjectMapParams(PipelineParams):
    accessor_id: str = Field(
        description='ID(s) of previously loaded accessor(s) to use as pipeline input'
    )
    pixel_classifier_segmentation_model_id: str = Field(
        description='Pixel classifier applied to segmentation_channel(s) to segment objects'
    )
    object_classifier_model_id: Union[str, None] = Field(
        None,
        description='Object classifier used to classify segmented objectss'
    )
    pixel_classifier_derived_model_id: Union[str, None] = Field(
        None,
        description='Pixel classifier used to derive channel(s) as additional inputs to object classification'
    )
    segmentation_channel: int = Field(
        description='Channel of input image to use for solving segmentation'
    )
    patches_channel: int = Field(
        description='Channel of input image used in patches sent to object classifier'
    )
    zmask_zindex: Union[int, None] = Field(
        None,
        description='z coordinate to use on input image when solving segmentation; apply MIP if empty',
    )
    roi_params: RoiSetMetaParams = RoiSetMetaParams(**{
        'mask_type': 'boxes',
        'filters': {
            'area': {'min': 1e3, 'max': 1e8}
        },
        'expand_box_by': [128, 2]
    })
    # TODO: maybe don't support all these exports here; instead leverage interm accessors
    export_params: RoiSetExportParams = RoiSetExportParams(**{
        'annotated_patches_2d': {
            'draw_bounding_box': True,
            'pad_to': 256,
        },
        'patches_2d': {
            'draw_bounding_box': True,
            'draw_mask': False,
        },
        'patch_masks': {},
        'object_classes': True,
        'dataframe': True,
    })
    derived_channels_input_channel: Union[int, None] = Field(
        None,
        description='Channel of input image from which to compute derived channels; use all if empty'
    )
    derived_channels_output_channels: Union[int, list] = Field(
        None,
        description='Derived channels to send to object classifier; use all if empty'
    )
    export_label_interm: bool = False


class RoiSetToObjectMapRecord(PipelineRecord):
    roiset_table: dict

@router.put('/roiset_to_obmap/infer')
def roiset_object_map(p: RoiSetObjectMapParams) -> RoiSetToObjectMapRecord:
    """
    Compute a RoiSet from 2d segmentation, apply to z-stack, and optionally apply object classification.
    """
    record, rois = call_pipeline(roiset_object_map_pipeline, p)

    # TODO: try labeling this correctly as input accessor_id instead of filename
    assert isinstance(rois, RoiSet)
    table = rois.get_serializable_dataframe()

    # TODO: instead, explicitly let trace include RoiSets
    session.write_to_table('RoiSet', {'input_filename': p.accessor_id}, table)
    ret = RoiSetToObjectMapRecord(
        roiset_table=table.to_dict(),
        **record.dict()
    )
    return ret


def roiset_object_map_pipeline(
        accessors: Dict[str, GenericImageDataAccessor],
        models: Dict[str, Model],
        **k
) -> (PipelineTrace, RoiSet):
    if not isinstance(models['pixel_classifier_segmentation_model'], SemanticSegmentationModel):
        raise IncompatibleModelsError('Expecting a pixel classification model')

    d = PipelineTrace(accessors['accessor'])
    zmi = k['zmask_zindex']
    sch = k['segmentation_channel']

    if isinstance(zmi, int):
        assert 0 < zmi < d.last.nz
        d['mip'] = d.last.get_mono(channel=sch).get_zi(zmi)
    else:
        d['mip'] = d.last.get_mono(channel=sch).apply(lambda x: x.max(axis=-1, keepdims=True))

    d['mip_mask'] = models['pixel_classifier_segmentation_model'].label_pixel_class(d.last)
    d['segmentation'] = get_label_ids(d.last)
    rois = RoiSet.from_object_ids(d['input'], d['segmentation'], RoiSetMetaParams(**k['roi_params']))

    # optionally run an object classifier if specified
    if obmod := models.get('object_classifier_model'):
        obmod_name = k['object_classifier_model_id']
        assert isinstance(obmod, InstanceSegmentationModel)
        rois.classify_by(
            obmod_name,
            [k['patches_channel']],
            obmod,
        )
        d[obmod_name] = rois.get_object_class_map(obmod_name)

    return d, rois

class Error(Exception):
    pass

class IncompatibleModelsError(Error):
    pass