from pathlib import Path
import unittest

import numpy as np


from model_server.base.accessors import generate_file_accessor
from tests.base.test_model import DummyInstanceSegmentationModel

import model_server.conf.testing as conf
from model_server.base.pipelines.roiset_obmap import RoiSetObjectMapParams, roiset_object_map_pipeline
from model_server.extensions.ilastik.models import IlastikPixelClassifierModel, IlastikPixelClassifierParams

data = conf.meta['image_files']
output_path = conf.meta['output_path']
test_params = conf.meta['roiset']
classifiers = conf.meta['ilastik_classifiers']


class BaseTestRoiSetMonoProducts(object):

    @property
    def fpi(self):
        return data['multichannel_zstack_raw']['path'].__str__()

    @property
    def fppx(self):
        return classifiers['px']['path'].__str__()

    @property
    def stack(self):
        return generate_file_accessor(self.fpi)

    @property
    def stack_ch_pa(self):
        return self.stack.get_mono(test_params['patches_channel'])

    @property
    def seg_mask(self):
        return generate_file_accessor(data['multichannel_zstack_mask2d']['path'])

    def _get_export_params(self):
        return {
            'pixel_probabilities': True,
            'patches_3d': {},
            'annotated_patches_2d': {
                'draw_bounding_box': True,
                'rgb_overlay_channels': [3, None, None],
                'rgb_overlay_weights': [0.2, 1.0, 1.0],
                'pad_to': 512,
            },
            'patches_2d': {
                'draw_bounding_box': False,
                'draw_mask': False,
            },
            'patch_masks': {
                'pad_to': 256,
            },
            'annotated_zstacks': {},
            'object_classes': True,
            'dataframe': True,
        }

    def _get_roi_params(self):
        return {
            'mask_type': 'boxes',
            'filters': {
                'area': {'min': 1e3, 'max': 1e8}
            },
            'expand_box_by': [128, 2]
        }

    def _get_models(self):
        return {
            'pixel_classifier_segmentation': {
                'name': 'ilastik_px_mod',
                'project_file': self.fppx,
                'model': IlastikPixelClassifierModel(
                    IlastikPixelClassifierParams(
                        project_file=self.fppx
                    )
                )
            },
            'object_classifier': {
                'name': 'dummy_ob_mod',
                'project_file': classifiers['seg_to_obj']['path'].__str__(),
                'model': DummyInstanceSegmentationModel()
            },
        }


class TestRoiSetWorkflow(BaseTestRoiSetMonoProducts, unittest.TestCase):

    def _pipeline_params(self):
        return {
            'api': False,
            'accessor_id': 'acc_id',
            'pixel_classifier_segmentation_model_id': 'px_id',
            'object_classifier_model_id': 'ob_id',
            'segmentation_channel': test_params['segmentation_channel'],
            'patches_channel': test_params['patches_channel'],
            'roi_params': self._get_roi_params(),
            'export_params': self._get_export_params(),
        }

    def test_object_map_workflow(self):
        acc_in = generate_file_accessor(self.fpi)
        params = RoiSetObjectMapParams(
            **self._pipeline_params(),
        )
        trace, _ = roiset_object_map_pipeline(
            {'accessor': acc_in},
            {f'{k}_model': v['model'] for k, v in self._get_models().items()},
            **params.dict()
        )

        self.assertTrue('ob_id' in trace.keys())
        self.assertEqual(len(trace['ob_id']._unique()[0]), 2)


    # def test_object_map_workflow_with_derived_channels(self):
    #     acc_in = generate_file_accessor(self.fpi)
    #     params = RoiSetObjectMapParams(
    #         **self._pipeline_params(),
    #         pixel_classifier_derived_model_id='id_px',
    #         derived_channels_input_channel=0,
    #         derived_channels_output_channels=[1],
    #     )
    #     models = self._get_models()
    #     models['pixel_classifier_derived'] = models['pixel_classifier_segmentation']  # re-use same classifier
    #
    #     trace, _ = roiset_object_map_pipeline(
    #         {'accessor': acc_in},
    #         {f'{k}_model': v['model'] for k, v in models.items()},
    #         **params.dict()
    #     )
    #
    #     self.assertTrue('ob_id' in trace.keys())
    #     self.assertEqual(len(trace['ob_id']._unique()[0]), 2)
    #     self.assertTrue('derived_00' in trace.keys())
    #     self.assertEqual(trace['derived_00'].chroma, 1)


class TestRoiSetWorkflowOverApi(conf.TestServerBaseClass, BaseTestRoiSetMonoProducts):

    input_data = data['multichannel_zstack_raw']


    def setUp(self) -> None:
        self.where_out = output_path / 'roiset'
        self.where_out.mkdir(parents=True, exist_ok=True)
        return conf.TestServerBaseClass.setUp(self)

    def test_trivial_api_response(self):
        resp = self._get('')
        self.assertEqual(resp.status_code, 200)

    def test_load_input_accessor(self):
        fname = self.copy_input_file_to_server()
        return self._put(f'accessors/read_from_file/{fname}').json()

    def test_load_pixel_classifier(self):
        resp = self._put(
            'ilastik/seg/load/',
            body={'project_file': self._get_models()['pixel_classifier_segmentation']['project_file']},
        )
        model_id = resp.json()['model_id']
        self.assertTrue(model_id.startswith('IlastikPixelClassifierModel'))
        return model_id

    def test_load_object_classifier(self):
        resp = self._put(
            'ilastik/seg_to_obj/load/',
            body={'project_file': self._get_models()['object_classifier']['project_file']},
        )
        model_id = resp.json()['model_id']
        self.assertTrue(model_id.startswith('IlastikObjectClassifierFromSegmentationModel'))
        return model_id

    def _object_map_workflow(self, ob_classifer_id):
        resp = self._put(
            'pipelines/roiset_to_obmap/infer',
            body={
                'accessor_id': self.test_load_input_accessor(),
                'pixel_classifier_segmentation_model_id': self.test_load_pixel_classifier(),
                'object_classifier_model_id': ob_classifer_id,
                'segmentation_channel': 0,
                'patches_channel': 1,
                'roi_params': self._get_roi_params(),
                'export_params': self._get_export_params(),
            },
        )
        self.assertEqual(resp.status_code, 200, resp.json())
        oid = resp.json()['output_accessor_id']
        obmap_fn = self._put(f'/accessors/write_to_file/{oid}').json()
        where_out = self._get('paths').json()['outbound_images']
        obmap_fp = Path(where_out) / obmap_fn
        self.assertTrue(obmap_fp.exists())
        return generate_file_accessor(obmap_fp)

    def test_workflow_with_object_classifier(self):
        acc = self._object_map_workflow(self.test_load_object_classifier())
        self.assertTrue(np.all(acc._unique()[0] == [0, 1]))

    def test_workflow_without_object_classifier(self):
        acc = self._object_map_workflow(None)
        self.assertTrue(np.all(acc._unique()[0] == [0, 255]))

    # TODO: test boundary pipeline in separate module
    # def test_object_map_workflow_boundary_channel(self):
    #     resp = self._put(
    #         'ilastik/seg/load/',
    #         body={
    #             'project_file': classifiers['px']['path'].__str__(),
    #             'px_class': 1,
    #             'px_prob_threshold': 0.5
    #         },
    #     )
    #     mid_px = resp.json()['model_id']
    #     mid_ob = self.test_load_object_classifier()
    #     resp = self._put(
    #         'chaeo/classify_zstack/infer',
    #         body=ClassifyZStackApiParams(
    #             model_ids={
    #                 'pixel_classifier_segmentation': mid_px,
    #                 'object_classifier': mid_ob,
    #             },
    #             **{
    #                 'input_filename': self.fpi,
    #                 'segmentation_channel': 0,
    #                 'patches_channel': 1,
    #                 'roi_params': _get_roi_params(),
    #                 'export_params': {'object_classes': True},
    #                 'label_params': {
    #                     'bbox_filter': {'area': {'min': 1e3, 'max': 1e8}},
    #                     'px_expand': 0,
    #                     'n_dilate': 2,
    #                     'marker_source': 'distmax',
    #                     'min_marker_dist': 10,
    #                     'background_method': 'threshold',
    #                     'coarse_sig': 15,
    #                     'background_tr': 0.2,
    #                 },
    #                 'export_label_interm': False,
    #             },
    #         ).dict()
    #     )
    #     self.assertEqual(resp.status_code, 200, resp.json())
    #     fp_obmap = Path(resp.json()['object_map_filepath'])
    #     print(fp_obmap)
    #     self.assertTrue(fp_obmap.exists())
    #     acc_obmap = generate_file_accessor(fp_obmap)
    #     self.assertTrue(all(np.unique(acc_obmap.data) == [0, 1]))


